package core.api.controller;

import core.api.models.auth.SimpleStatusResponse;
import core.api.models.timemanager.tasks.*;
import core.entities.timemanager.Task;
import core.entities.timemanager.Taskgroup;
import core.services.*;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.ResponseEntity;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.web.bind.annotation.*;

import javax.validation.Valid;
import java.util.ArrayList;
import java.util.List;

@CrossOrigin(origins = "*", maxAge = 3600)
@RestController
@RequestMapping("/api")
public class TaskController {

    private final TaskService taskService;
    private final TaskgroupService taskgroupService;

    public TaskController(@Autowired TaskService taskService, @Autowired TaskgroupService taskgroupService) {
        this.taskService = taskService;
        this.taskgroupService = taskgroupService;
    }

    @GetMapping("/tasks/all/{scope}/{detailed}")
    public ResponseEntity<?> loadAllTasksShort(@PathVariable TaskScope scope, @PathVariable boolean detailed) {
        List<Task> taskList = taskService.loadAllTasks(SecurityContextHolder.getContext().getAuthentication().getName(), scope);
        List<TaskShortInfo> taskShortInfos = new ArrayList<>();
        List<TaskTaskgroupInfo> taskInfos = new ArrayList<>();
        for(Task task : taskList) {
            StringBuilder builder = loadTaskgroupPath(task);
            if(!detailed) {
                taskShortInfos.add(new TaskShortInfo(task.getTaskID(), builder.toString()));
            } else {
                taskInfos.add(new TaskTaskgroupInfo(task));
            }

        }

        if(detailed) {
            return ResponseEntity.ok(taskInfos);
        } else {
            return ResponseEntity.ok(taskShortInfos);
        }
    }

    private StringBuilder loadTaskgroupPath(Task task) {
        List<Taskgroup> ancestors = Taskgroup.getAncestorList(task.getTaskgroup());
        ancestors.add(task.getTaskgroup());
        StringBuilder builder = new StringBuilder();
        for(Taskgroup taskgroup : ancestors) {
            builder.append(taskgroup.getTaskgroupName()).append("/");
        }
        builder.append(task.getTaskName());
        return builder;
    }

    @GetMapping("/tasks/{taskgroupID}/{status}")
    public ResponseEntity<?> listTasksOfTaskgroup(@PathVariable long taskgroupID, @PathVariable String status) {
        PermissionResult<Taskgroup> taskgroupPermissionResult = taskgroupService.getTaskgroupByIDAndUsername(taskgroupID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!taskgroupPermissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(taskgroupPermissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        return ResponseEntity.ok(taskgroupPermissionResult.getResult().getTasks().stream().map(TaskEntityInfo::new));
    }

    @PutMapping("/tasks/{taskgroupID}")
    public ResponseEntity<?> createTask(@PathVariable long taskgroupID, @RequestBody @Valid TaskFieldInfo taskFieldInfo) {
        PermissionResult<Taskgroup> taskgroupPermissionResult = taskgroupService.getTaskgroupByIDAndUsername(taskgroupID, SecurityContextHolder.getContext().getAuthentication().getName());
        if(!taskgroupPermissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if(taskgroupPermissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        ServiceResult<Task> creationResult = taskService.createTask(taskgroupPermissionResult.getResult(), taskFieldInfo);
        if(creationResult.getExitCode() == ServiceExitCode.ENTITY_ALREADY_EXIST) {
            return ResponseEntity.status(409).body(new SimpleStatusResponse("failed"));
        } else {
            return ResponseEntity.ok(new TaskEntityInfo(creationResult.getResult()));
        }
    }

    @PostMapping("/tasks/{taskID}")
    public ResponseEntity<?> editTask(@PathVariable long taskID, @RequestBody @Valid TaskFieldInfo taskFieldInfo) {
        PermissionResult<Task> taskPermissionResult = taskService.getTaskPermissions(taskID, SecurityContextHolder.getContext().getAuthentication().getName());
        if (!taskPermissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if (taskPermissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        if(taskFieldInfo.getStartDate().isAfter(taskFieldInfo.getDeadline())) {
            return ResponseEntity.status(400).body(new SimpleStatusResponse("failed"));
        }

        ServiceResult<Task> editorResult = taskService.editTask(taskPermissionResult.getResult(), taskFieldInfo);
        if (editorResult.getExitCode() == ServiceExitCode.OK) {
            return ResponseEntity.ok(new SimpleStatusResponse("success"));
        } else  {
            return ResponseEntity.status(409).body(new SimpleStatusResponse("failed"));
        }
    }

    @DeleteMapping("/tasks/{taskID}")
    public ResponseEntity<?> deleteTask(@PathVariable long taskID) {
        PermissionResult<Task> taskPermissionResult = taskService.getTaskPermissions(taskID, SecurityContextHolder.getContext().getAuthentication().getName());
        if (!taskPermissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if (taskPermissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        taskService.deleteTask(taskPermissionResult.getResult());
        return ResponseEntity.ok(new SimpleStatusResponse("success"));
    }

    @GetMapping("/tasks/{taskID}")
    public ResponseEntity<?> loadTaskDetails(@PathVariable long taskID) {
        PermissionResult<Task> taskPermissionResult = taskService.getTaskPermissions(taskID, SecurityContextHolder.getContext().getAuthentication().getName());
        if (!taskPermissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if (taskPermissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        return ResponseEntity.ok(new TaskEntityInfo(taskPermissionResult.getResult()));
    }

    @PostMapping("/tasks/{taskID}/finish")
    public ResponseEntity<?> finishTask(@PathVariable long taskID) {
        PermissionResult<Task> taskPermissionResult = taskService.getTaskPermissions(taskID, SecurityContextHolder.getContext().getAuthentication().getName());
        if (!taskPermissionResult.isHasPermissions()) {
            return ResponseEntity.status(403).body(new SimpleStatusResponse("failed"));
        }

        if (taskPermissionResult.getExitCode() == ServiceExitCode.MISSING_ENTITY) {
            return ResponseEntity.status(404).body(new SimpleStatusResponse("failed"));
        }

        taskService.finishTask(taskPermissionResult.getResult());
        return ResponseEntity.ok(new SimpleStatusResponse("success"));
    }
}
